# see how the GIL bites us in the behind...
# example by Charl P. Botha <http://cpbotha.net/> 2005

# testing this on Python 2.3 on Windows XP and Python 2.4 on Linux x86,
# one sees that the VTK pipelines are executed sequentially, although
# they are in separate threads.

# I *could* make them run in parallel using VTK-specific calls or by doing
# funky things in VTK progress handlers, but what I REALLY want is a generic 
# approach to getting arbitrary wrapped (long-running) code run in 
# parallel threads from Python.  The GIL is preventing me from doing this.  
# Thanks GIL.

import Queue
import threading
import time

import vtk

class ModuleManager(object):
    def __init__(self):
        self.queue = Queue.Queue(0)

    def run(self):
        # create modules and threads
        modules = []
        threads = []

        for i in range(4):
            modules.append(moduleBase(self))
            threads.append(threading.Thread(target=modules[-1].run))
            threads[-1].start()

        # send messages to get them started
        for m in modules:
            m.queue.put("do something")

        # listen for incoming events
        try:
            curm = None
            while 1:
                o,m = self.queue.get(block=True, timeout=5)
                if o != curm:
                    curm = o
                    print (o, m)
                    
        except Queue.Empty:
            pass

        # tell all modules to stop by sending them this message
        for m in modules:
            m.queue.put("stop")

        # wait for them to join
        for t in threads:
            t.join()

class moduleBase(object):
    def __init__(self, module_manager):
        self._module_manager = module_manager
        # message queue of this module, size infinite
        self.queue = Queue.Queue(0)


    def run(self):
        """This will be run in a thread.  It has to wait for messages from
        our queue and react.
        """

        msg = None
        while msg != "stop":
            msg = self.queue.get(block=True)

            if msg != "stop":
                self._r = vtk.vtkXMLImageDataReader()
                self._r.SetFileName('/home/cpbotha/work/data1/volvis.org/aneurism.vti')
                self._v = vtk.vtkMarchingCubes()
                self._v.SetInput(self._r.GetOutput())
                self._v.SetValue(0, 201)
                print "starting update thread ", self, "\n"
                self._v.Update()
                print "done thread ", self, "\n"
                self._module_manager.queue.put((
                    self, 'done'))

if __name__ == '__main__':

    a = ModuleManager()
    a.run()
